//---------------------------------------------------------------------------
// GPGee - GNU Privacy Guard Explorer Extension
// GPGeeSignEncrypt.cpp - 'Main' form code
// Copyright 2005, Kurt Fitzner <kfitzner@excelcia.org>
//---------------------------------------------------------------------------
// This file is part of GPGee.
//
// GPGee is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// GPGee is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
//---------------------------------------------------------------------------

//---------------------------------------------------------------------------
// File Notes:
// This is the code behind the sign/encrypt form.  It's often not
// encouraged to put user interface code in the same place as functional
// code.  I've never subscribed to this seperation, at least not for smaller
// applications.  I find it decreases code readability when a user interface
// function callback for a button click event does nothing but turn around
// and call something else.  You will find most of the functional code
// right in the event handler callback functions here.
//---------------------------------------------------------------------------
#include <vcl.h>
#include <Registry.hpp>
#pragma hdrstop

#include "TConfiguration.h"
#include "GPGeeExceptions.h"
#include "GPGeePassPhrase.h"
#include "GPGeeUtility.h"
#include "GPGeeSignEncrypt.h"
//---------------------------------------------------------------------------
#pragma package(smart_init)
#pragma resource "*.dfm"
TformGPGeeSignEncrypt *formGPGeeSignEncrypt = NULL;
extern TConfiguration *GPGeeConfig;
//---------------------------------------------------------------------------

//---------------------------------------------------------------------------
// Constructor for TformGPGeeSignEncrypt - our sign/encryption form
//
__fastcall TformGPGeeSignEncrypt::TformGPGeeSignEncrypt(TComponent* Owner) : TForm(Owner)
{
  gpgme_error_t err;
  char *buffer;
  AnsiString sConfigString;

  // A flag to signal that when the form is activated after contstruction that it is the first time.
  //bFirstActivation = true;

  ctxGPG = NULL;
  cachePublic = NULL;
  cacheSecret = NULL;

  // Initialize the GPGME library main and keycache contexts
  err = gpgme_new(&ctxGPG);
  if (err != GPGME_No_Error)
    throw EGPGMEError(FormMessage(MSG_ERROR_GPGME_LIB_INIT, gpgme_strerror(err)));
  err = gpgme_keycache_new (&cachePublic, ctxGPG);
  if (err != GPGME_No_Error)
    throw EGPGMEError(FormMessage(MSG_ERROR_GPGME_KEYCACHE_INIT_PUB,gpgme_strerror(err)));
  err = gpgme_keycache_new (&cacheSecret, ctxGPG);
  if (err != GPGME_No_Error)
    throw EGPGMEError(FormMessage(MSG_ERROR_GPGME_KEYCACHE_INIT_SEC,gpgme_strerror(err)));

  // Set GPGME's settings to match GPGee's configuration
  gpgme_control(ctxGPG, GPGME_CTRL_FORCETRUST, GPGeeConfig->Values["Force Trust"]?1:0);
  sConfigString = GPGeeConfig->Values["GPG Program"];
  if (!sConfigString.IsEmpty())
    gpgme_control(ctxGPG, GPGME_CTRL_GPGPROGRAM, (int)sConfigString.c_str());
  sConfigString = GPGeeConfig->Values["Options File"];
  if (!sConfigString.IsEmpty())
    gpgme_control(ctxGPG, GPGME_CTRL_OPTFILE, (int)sConfigString.c_str());
  sConfigString = GPGeeConfig->Values["Keyring"];
  if (!sConfigString.IsEmpty())
    gpgme_control(ctxGPG, GPGME_CTRL_KEYRING, (int)sConfigString.c_str());
  sConfigString = GPGeeConfig->Values["Secret Keyring"];
  if (!sConfigString.IsEmpty())
    gpgme_control(ctxGPG, GPGME_CTRL_SECRETKEYRING, (int)sConfigString.c_str());

  pubkeysLastSorted = -1;
  pubkeysAscending = true;
  nEncryptCount = 0;
}  // __fastcall TformGPGeeSignEncrypt::TformGPGeeSignEncrypt(TComponent* Owner) : TForm(Owner)
//---------------------------------------------------------------------------


//---------------------------------------------------------------------------
// Destructor
//
__fastcall TformGPGeeSignEncrypt::~TformGPGeeSignEncrypt()
{
  if (cacheSecret)
    gpgme_keycache_release(cacheSecret);
  if (cachePublic)
    gpgme_keycache_release(cachePublic);
  if (ctxGPG)
    gpgme_release(ctxGPG);
}  // __fastcall TformGPGeeSignEncrypt::~TformGPGeeSignEncrypt()
//---------------------------------------------------------------------------


//---------------------------------------------------------------------------
// Set the default modes and show the form - this just stores the default
// modes.  The real work of actually changing the controls on the form to
// match these defaults is done in FormShow()
//
int __fastcall TformGPGeeSignEncrypt::ShowModalWithDefaults(bool bSign, bool bEncrypt, TStringList *FileNames)
{
  bStartupSign = bSign;
  bStartupEncrypt = bEncrypt;
  this->FileNames = FileNames;
  return ShowModal();
}  // int __fastcall TformGPGeeSignEncrypt::ShowModalWithDefaults(bool bSign, bool bEncrypt)
//---------------------------------------------------------------------------


//---------------------------------------------------------------------------
// OnShow event handler - The OnShow event occurs before the form becomes
// visible, but after all its components have been streamed in.  This is a
// good place to put code that initializes the form's look so that the form
// looks like it should right when it first pops up.
//
void __fastcall TformGPGeeSignEncrypt::FormShow(TObject *Sender)
{
  gpgme_error_t err;
  gpgme_key_t key;
  TListItem *keyvisual;
  const char *caps;
  int nSubkey;
  bool bValidkey;
  AnsiString sValidity;
  int nDefaultSigningKey = GPGeeConfig->Values["Signing Key"];

  // Create one keycache each for the public and private keys
  err = gpgme_keycache_init(cachePublic, NULL, false);
  if (err)
    throw EGPGMEError(FormMessage(MSG_ERROR_GPGME_KEYCACHE_LOAD_PUB,gpgme_strerror(err)));
  err = gpgme_keycache_init(cacheSecret, NULL, true);
  if (err)
    throw EGPGMEError(FormMessage(MSG_ERROR_GPGME_KEYCACHE_LOAD_SEC,gpgme_strerror(err)));

  // Fill in the info on keys that can be encrypted to
  lvPublicKeys->Items->Clear();
  gpgme_keycache_rewind(cachePublic);
  while (gpgme_keycache_next_key(cachePublic, false, &key) == 0) {
    // If the key can be encrypted to, add its info to the listview component
    nSubkey = 0;
    bValidkey = false;
    // Find the first subkey that can be encrypted to
    while ((caps = gpgme_key_get_string_attr(key, GPGME_ATTR_KEY_CAPS, NULL, nSubkey)) != NULL) {
      // If the key can be encrypted to and isn't revoked or expired, we'll call it valid and add it
      if (strchr(caps, 'e') && gpgme_key_get_ulong_attr(key, GPGME_ATTR_KEY_USABLE, NULL, nSubkey)) {
        bValidkey = true;
        break;
      }  // if (strchr(caps, 'e'))
      nSubkey++;
    }  // while ((caps = gpgme_key_get_string_attr(key, GPGME_ATTR_KEY_CAPS, NULL, nSubkey)) != NULL)
    if (!bValidkey)
      continue;
    // Add a new item to the listview component and then add all the information to it for each column.
    keyvisual = lvPublicKeys->Items->Add();
    keyvisual->Caption = String(gpgme_key_get_string_attr(key, GPGME_ATTR_NAME, NULL, 0)) + " <" + String(gpgme_key_get_string_attr(key, GPGME_ATTR_EMAIL, NULL, 0)) + ">";
    keyvisual->SubItems->Add("0x" + String(gpgme_key_get_string_attr(key, GPGME_ATTR_KEYID, NULL, nSubkey)).Delete(1,8));
    keyvisual->SubItems->Add(String(gpgme_key_get_ulong_attr(key, GPGME_ATTR_LEN, NULL, nSubkey)));
    keyvisual->SubItems->Add(String(gpgme_key_get_string_attr(key, GPGME_ATTR_ALGO, NULL, nSubkey)));
    switch (gpgme_key_get_ulong_attr(key, GPGME_ATTR_VALIDITY, NULL, 0)) {
      case GPGME_VALIDITY_UNKNOWN:   sValidity = GetMessage(MSG_GPGME_VALIDITY_UNKNOWN); break;
      case GPGME_VALIDITY_UNDEFINED: sValidity = GetMessage(MSG_GPGME_VALIDITY_UNDEFINED); break;
      case GPGME_VALIDITY_NEVER:     sValidity = GetMessage(MSG_GPGME_VALIDITY_NEVER); break;
      case GPGME_VALIDITY_MARGINAL:  sValidity = GetMessage(MSG_GPGME_VALIDITY_MARGINAL); break;
      case GPGME_VALIDITY_FULL:      sValidity = GetMessage(MSG_GPGME_VALIDITY_FULL); break;
      case GPGME_VALIDITY_ULTIMATE:  sValidity = GetMessage(MSG_GPGME_VALIDITY_ULTIMATE); break;
      default:                       sValidity = GetMessage(MSG_GPGME_VALIDITY_ERROR); break;
    }  // switch (gpgme_key_get_ulong_attr(key, GPGME_ATTR_VALIDITY, NULL, 0))
    keyvisual->SubItems->Add(sValidity);
  }  // while (gpgme_op_keylist_next(ctxGPG, &key) == 0)

  cbSigningKey->Clear();
  gpgme_keycache_rewind(cacheSecret);
  // Add the secret keys to the signing key selection drop-down combobox.
  while (gpgme_keycache_next_key(cacheSecret, false, &key) == 0) {
    AnsiString sName = String(gpgme_key_get_string_attr(key, GPGME_ATTR_NAME, NULL, 0)) + " <" + String(gpgme_key_get_string_attr(key, GPGME_ATTR_EMAIL, NULL, 0)) + ">";
    AnsiString sAlgo = String(gpgme_key_get_string_attr(key, GPGME_ATTR_ALGO, NULL, 0));
    AnsiString sKeyId = "0x" + String(gpgme_key_get_string_attr(key, GPGME_ATTR_KEYID, NULL, 0)).Delete(1,8);
    unsigned nKeyId = StrToInt(sKeyId);
    cbSigningKey->Items->AddObject(sName + " (" + sAlgo + "/" + sKeyId + ")", (TObject *)nKeyId);
  }  // while (gpgme_op_keylist_next(ctxGPG, &key) == 0)
  // Find the default signing key (if any) and select it
  cbSigningKey->ItemIndex = 0;
  for (int n = 0; n < cbSigningKey->Items->Count; n++)
    if ((int)cbSigningKey->Items->Objects[n] == nDefaultSigningKey)
      cbSigningKey->ItemIndex = n;

  // Set the form up with the defaults we were given for signing/encryption in ShowModalWithDefaults()
  if (bStartupSign && !bStartupEncrypt) {
    radioDetached->Checked = true;
    chbASCIIArmor->Checked = true;
  } else if (!bStartupSign && bStartupEncrypt) {
    rgEncryptionOptions->ItemIndex = GPGEE_ENCRYPTION_PUBLIC;
  } else if (bStartupSign && bStartupEncrypt) {
    rgEncryptionOptions->ItemIndex = GPGEE_ENCRYPTION_PUBLIC;
    radioAttached->Checked = true;
  }  // default signing options
  SanityCheck(NULL);
}  // void __fastcall TformGPGeeSignEncrypt::FormActivateFirst(TObject *Sender)
//---------------------------------------------------------------------------


//---------------------------------------------------------------------------
// OnColumnClick event handler for the public keys listview component
// Change the column that the list is sorted by
//
void __fastcall TformGPGeeSignEncrypt::lvPublicKeysColumnClick(TObject *Sender, TListColumn *Column)
{
  // If a column is clicked on once, sort by that column ascending - if it is clicked on again before another
  // column is clicked on, then sort by that column descending.
  if (Column->Index == pubkeysLastSorted)
    pubkeysAscending = !pubkeysAscending;
  else
    pubkeysLastSorted = Column->Index;
  lvPublicKeys->CustomSort(SortByColumn, Column->Index);
}  // void __fastcall TformGPGeeSignEncrypt::lvPublicKeysColumnClick(TObject *Sender, TListColumn *Column)
//---------------------------------------------------------------------------


//---------------------------------------------------------------------------
// Custom listview sort comparison callback function
// Sort by the column selected by the user
//
int CALLBACK SortByColumn(long Item1, long Item2, long Data)
{
  int retval;

  if (!Data)
    retval = AnsiCompareText(((TListItem *)Item1)->Caption, ((TListItem *)Item2)->Caption);
  else
    retval = AnsiCompareText(((TListItem *)Item1)->SubItems->Strings[Data-1], ((TListItem *)Item2)->SubItems->Strings[Data-1]);

  if (!pubkeysAscending)
    retval = -retval;

  return retval;
}  // int __stdcall SortByColumn(long Item1, long Item2, long Data)
//---------------------------------------------------------------------------


//---------------------------------------------------------------------------
// Perform a sanity check on the settings and (attempt to) intelligently
// alter any that don't make sense.  This is where all the logic is that
// determines what controls are active depending on what options are
// selected.
//
void __fastcall TformGPGeeSignEncrypt::SanityCheck(TObject *Sender)
{
  // Form defaults that depend on the state of the Encryption Options radio group box. The state of the Encryption
  // Options affects what sort of signing options are available and whether the public key list view is enabled.  The
  // listview isn't actually disabled since it looks gross that way.  We fake disabling it by greying out the text and
  // not allowing any changes to it (see the lvPublicKeysChanging() method)
  switch (rgEncryptionOptions->ItemIndex) {
    case GPGEE_ENCRYPTION_NONE:
      if (lvPublicKeys->Font->Color == clWindowText) {
        lvPublicKeys->MultiSelect = false;
        lvPublicKeys->Selected = NULL;
        lvPublicKeys->MultiSelect = true;
        lvPublicKeys->Font->Color = clInactiveCaptionText;
      }  // if (lvPublicKeys->Font->Color == clWindowText)
      radioNoSignature->Enabled = true;
      radioAttached->Enabled = true;
      radioDetached->Enabled = true;
      break;
    case GPGEE_ENCRYPTION_PUBLIC:
      lvPublicKeys->Font->Color = clWindowText;
      radioNoSignature->Enabled = true;
      radioAttached->Enabled = true;
      if (radioDetached->Checked)
        radioAttached->Checked = true;
      radioDetached->Enabled = false;
      break;
    case GPGEE_ENCRYPTION_SYMMETRIC:
      if (lvPublicKeys->Font->Color == clWindowText) {
        lvPublicKeys->MultiSelect = false;
        lvPublicKeys->Selected = NULL;
        lvPublicKeys->MultiSelect = true;
        lvPublicKeys->Font->Color = clInactiveCaptionText;
      }  // if (lvPublicKeys->Font->Color == clWindowText)
      radioNoSignature->Enabled = false;
      radioAttached->Enabled = false;
      radioDetached->Enabled = false;
      break;
  }  // switch (rgEncryptionOptions->ItemIndex)

  cbSigningKey->Enabled = !radioNoSignature->Checked && radioNoSignature->Enabled;
  chbClearsign->Enabled = radioAttached->Checked && radioAttached->Enabled  && rgEncryptionOptions->ItemIndex != GPGEE_ENCRYPTION_PUBLIC;
  chbASCIIArmor->Enabled = rgEncryptionOptions->ItemIndex || !radioNoSignature->Checked;
  btnOk->Enabled = rgEncryptionOptions->ItemIndex==GPGEE_ENCRYPTION_SYMMETRIC ||  rgEncryptionOptions->ItemIndex==GPGEE_ENCRYPTION_PUBLIC && nEncryptCount || !rgEncryptionOptions->ItemIndex && !radioNoSignature->Checked;
  if (radioNoSignature->Enabled && !radioNoSignature->Checked && rgEncryptionOptions->ItemIndex != GPGEE_ENCRYPTION_NONE)
    IconList->GetIcon(0, Icon);
  else if (radioNoSignature->Enabled && !radioNoSignature->Checked && rgEncryptionOptions->ItemIndex == GPGEE_ENCRYPTION_NONE)
    IconList->GetIcon(1, Icon);
  else
    IconList->GetIcon(2, Icon);

}  // void __fastcall TformGPGeeSignEncrypt::SanityCheck(TObject *Sender)
//---------------------------------------------------------------------------


//---------------------------------------------------------------------------
// TListView OnChanging event handler
// Forbid any changes if the text color is clInactiveCaptionText
// We use the text color as the indicator of whether the TListView component
// is enabled or not.  This is instead of using the Enabled property because
// setting Enabled to false makes the listview look really really ugly.
// Graying out the text is much more professional looking.
//
void __fastcall TformGPGeeSignEncrypt::lvPublicKeysChanging(TObject *Sender, TListItem *Item, TItemChange Change, bool &AllowChange)
{
  AllowChange = lvPublicKeys->Font->Color != clInactiveCaptionText;
}
//---------------------------------------------------------------------------


//---------------------------------------------------------------------------
// TListView OnChange event handler
// We use this event as a signal to count how many keys are checkmarked.
// This count is used so we know when to intelligently enable the Ok button
//
void __fastcall TformGPGeeSignEncrypt::lvPublicKeysChange(TObject *Sender, TListItem *Item, TItemChange Change)
{
  int nCount = 0;
  int nOldEncryptCount = nEncryptCount;

  for (int n = 0; n < lvPublicKeys->Items->Count; n++)
    if (lvPublicKeys->Items->Item[n]->Checked)
      nCount++;
  nEncryptCount = nCount;
  if (nEncryptCount != nOldEncryptCount)
    SanityCheck(NULL);
}  // void __fastcall TformGPGeeSignEncrypt::lvPublicKeysChange(TObject *Sender, TListItem *Item, TItemChange Change)
//---------------------------------------------------------------------------


//---------------------------------------------------------------------------
// Help button OnClick event handler
// Bring up the help file on the title page
void __fastcall TformGPGeeSignEncrypt::btnHelpClick(TObject *Sender)
{
  Application->HelpContext(2);
}
//---------------------------------------------------------------------------


//---------------------------------------------------------------------------
// btnOk OnClick handler
// Time to roll up our forearms and work some magic.  This is where the
// actual work of signing/encryption is done.
//
void __fastcall TformGPGeeSignEncrypt::btnOkClick(TObject *Sender)
{
  AnsiString sSigningKeyId, sSigningKeyName, sSigningKeyType, sSigningKeySize, sSigningKeyCreationDate;
  gpgme_key_t keySecret;
  gpgme_recipients_t keyRecipients = NULL;
  gpgme_error_t err;
  AnsiString sPassphrase;
  int nPassphraseResult;
  gpgme_sigmode_t nSignatureMode;

  formGPGeePassPhrase = NULL;
  try {
    formGPGeePassPhrase = new TformGPGeePassPhrase(NULL);
    // If we've been asked to sign anything, then we're going to need to grab the signing key id and then get a passphrase
    // We also set the signature mode here.
    if (!radioNoSignature->Checked && radioNoSignature->Enabled) {
      sSigningKeyId = IntToHex((int)cbSigningKey->Items->Objects[cbSigningKey->ItemIndex],8);
      gpgme_keycache_rewind(cacheSecret);
      err = gpgme_keycache_find_key(cacheSecret, sSigningKeyId.c_str(), false, &keySecret);
      if (err)
        throw EGPGMEError(FormMessage(MSG_ERROR_GPGEE_FIND_SECRETKEY,sSigningKeyId.c_str(),gpgme_strerror(err)));
      // Grab all the info about the signing key to provide the user in the passphrase dialog
      sSigningKeyName = String(gpgme_key_get_string_attr(keySecret, GPGME_ATTR_NAME, NULL, 0)) + " <" + String(gpgme_key_get_string_attr(keySecret, GPGME_ATTR_EMAIL, NULL, 0)) + ">";
      sSigningKeyType = String(gpgme_key_get_string_attr(keySecret, GPGME_ATTR_ALGO, NULL, 0));
      sSigningKeySize = String(gpgme_key_get_ulong_attr(keySecret, GPGME_ATTR_LEN, NULL, 0));
      sSigningKeyCreationDate = DateToStr(GPGToDateTime(gpgme_key_get_ulong_attr(keySecret, GPGME_ATTR_CREATED, NULL, 0)));
      // Get the passphrase
      nPassphraseResult = formGPGeePassPhrase->GetPassphrase(sSigningKeyName, sSigningKeyId, sSigningKeyType, sSigningKeySize, sSigningKeyCreationDate);
      if (nPassphraseResult == mrOk)
        sPassphrase = formGPGeePassPhrase->edtPassphrase->Text;
      else if (nPassphraseResult == mrCancel)
        return;
      if (radioAttached->Checked)
        if (chbClearsign->Checked && chbClearsign->Enabled)
          nSignatureMode = GPGME_SIG_MODE_CLEAR;
        else
          nSignatureMode = GPGME_SIG_MODE_NORMAL;
      else
        nSignatureMode = GPGME_SIG_MODE_DETACH;
      gpgme_signers_clear(ctxGPG);
      err = gpgme_signers_add(ctxGPG, keySecret);
      if (err)
        throw EGPGMEError(FormMessage(MSG_ERROR_GPGEE_SET_SIGNINGKEY,sSigningKeyId.c_str(),gpgme_strerror(err)));
    }  // if (!radioNoSignature->Checked)

    // Check if we are encrypting and set things up if we are...
    if (rgEncryptionOptions->ItemIndex == 1) {
      gpgme_recipients_new(&keyRecipients);
      for (int n = 0; n < lvPublicKeys->Items->Count; n++) {
        TListItem *item = lvPublicKeys->Items->Item[n];
        if (item->Checked) {
          err = gpgme_recipients_add_name(keyRecipients, item->SubItems->Strings[0].c_str());
          if (err)
            throw EGPGMEError(FormMessage(MSG_ERROR_GPGEE_ADD_RECIPIENT,item->SubItems->Strings[0].c_str(),gpgme_strerror(err)));
        }  // if (item->Checked)
      }  // for (int n = 0; n < lvPublicKeys->Items->Count; n++)
    } else if (rgEncryptionOptions->ItemIndex == GPGEE_ENCRYPTION_SYMMETRIC) {
      nPassphraseResult = formGPGeePassPhrase->GetPassphrase("", "", "", "", "");
      if (nPassphraseResult == mrOk)
        sPassphrase = formGPGeePassPhrase->edtPassphrase->Text;
      else
        return;
    }  // if (rgEncryptionOptions->ItemIndex == 1)

    gpgme_control(ctxGPG, GPGME_CTRL_ARMOR, chbASCIIArmor->Checked);
    // Whatever we've just been told to do, do it for each file in the list...
    for (int n = ComponentCount - 1; n >= 0; n--)
      if (Components[n]->InheritsFrom(__classid(TControl)))
        ((TControl *)Components[n])->Cursor = crHourGlass;
    Application->ProcessMessages();
    for (int n=0; n < FileNames->Count; n++) {
      AnsiString FileName = "\"" + FileNames->Strings[n] + "\"";
      switch (rgEncryptionOptions->ItemIndex) {
        case GPGEE_ENCRYPTION_NONE:
          gpgme_set_passphrase(ctxGPG, sPassphrase.c_str());
          err = gpgme_op_file_sign(ctxGPG, nSignatureMode, FileName.c_str(), NULL);
          break;
        case GPGEE_ENCRYPTION_PUBLIC:
          // Check if we are signing and encrypting, or encrypting only
          if (radioNoSignature->Enabled && !radioNoSignature->Checked) {
            // We are signing and encrypting, so we need to set the passphrase
            gpgme_set_passphrase(ctxGPG, sPassphrase.c_str());
            // if encrypt-to-self is set then add the key used to sign to the list of recipients
            if (GPGeeConfig->Values["Encrypt to Self"]) {
              gpgme_key_t EncryptSelfKey;
              int nSubkey;
              const char *caps;
              bool bValidkey = false;

              // Find the encryption subkey of the key that is used to sign with...
              gpgme_keycache_rewind(cachePublic);
              err = gpgme_keycache_find_key(cachePublic, sSigningKeyId.c_str(), false, &EncryptSelfKey);
              if (err != GPGME_No_Error) {
                while ((caps = gpgme_key_get_string_attr(EncryptSelfKey, GPGME_ATTR_KEY_CAPS, NULL, nSubkey)) != NULL) {
                  // If the key can be encrypted to and isn't revoked or expired, we'll call it valid and add it
                  if (strchr(caps, 'e') && gpgme_key_get_ulong_attr(EncryptSelfKey, GPGME_ATTR_KEY_USABLE, NULL, nSubkey)) {
                    bValidkey = true;
                    break;
                  }  // if (strchr(caps, 'e'))
                  nSubkey++;
                }  // while ((caps = gpgme_key_get_string_attr(key, GPGME_ATTR_KEY_CAPS, NULL, nSubkey)) != NULL)
                if (bValidkey)
                  gpgme_recipients_add_name(keyRecipients, gpgme_key_get_string_attr(EncryptSelfKey, GPGME_ATTR_KEYID, NULL, nSubkey));
              }  // if (err != GPGME_No_Error)
            }  // if (GPGeeConfig->Values["Encrypt to Self"])
            // Sign and encrypt the file
            err = gpgme_op_file_sign_encrypt(ctxGPG, keyRecipients, FileName.c_str(), NULL);
          } else
            // Encrypt the file only
            err = gpgme_op_file_encrypt(ctxGPG, keyRecipients, FileName.c_str(), NULL);
          break;
        case GPGEE_ENCRYPTION_SYMMETRIC:
          // We are performing symmetrical encryption so we need to set the passphrase
          gpgme_set_passphrase(ctxGPG, sPassphrase.c_str());
          err = gpgme_op_file_encrypt(ctxGPG, NULL, FileName.c_str(), NULL);
      }  // switch (rgEncryptionOptions->ItemIndex)
      if (err)
        // If the operation fails, tell the user about it and bail out
        throw EGPGMEError(FormMessage(MSG_ERROR_GPGEE_FAILED_OPERATION,gpgme_strerror(err)));
    }  // for (int n=0; n < FileNames->Count; n++)
    // At this point we have successfully performed the operation - if we have signed a file, then set the key we used
    // to sign with as the default signing key for next time.
    if (radioNoSignature->Enabled && !radioNoSignature->Checked)
      GPGeeConfig->Values["Signing Key"] = (int)cbSigningKey->Items->Objects[cbSigningKey->ItemIndex];
    ModalResult = mrOk;
  }  // try

  // Cleanup after ourselves
  __finally {
    // Set the cursor back to default - in case there was an error and the form is still up.
    for (int n = ComponentCount - 1; n >= 0; n--)
      if (Components[n]->InheritsFrom(__classid(TControl)))
        ((TControl *)Components[n])->Cursor = crDefault;
    // Free allocated memory for our recipient key list (if any)
    if (keyRecipients)
      gpgme_recipients_release(keyRecipients);
    // Here we used to wipe the passphrase copy that was in the passphrase dialog but this is now done in the dialog's destructor
    if (formGPGeePassPhrase)
      delete formGPGeePassPhrase;
    // Wipe the local copy of the passphrase we used and also the copy we sent to gpgme.  It's possible the passphrase could
    // still end up on a swap file, but this is unlikely and preventing that would add huge complexity.  This will have to suffice.
    sPassphrase = sPassphrase.StringOfChar('*',sPassphrase.Length());
    gpgme_set_passphrase(ctxGPG, sPassphrase.c_str());
  }  // __finally
}  // void __fastcall TformGPGeeSignEncrypt::btnOkClick(TObject *Sender)
//---------------------------------------------------------------------------

